import { Point } from "@influxdata/influxdb-client";
import { Request, Response } from "express";
import Cache from "node-cache";
import { Route, RouteMessages } from "server";
import { InfluxWriteApi, nodeManager } from "..";
import { MetricsSchema } from "../db/metrics";

const IGNORED_PATHS = ["/favicon.ico"];

/**
 * The total requests that have been made
 * TODO: move this to a metrics file
 */
let totalRequests: number | undefined = undefined;

const cache = new Cache({
  stdTTL: 300, // 5 minutes
});
type CachedRequest = {
  nodeId: string;
  status: number;
  headers: any;
  data: any;
};

/**
 * Logs a request
 *
 * @param nodeId the node ID that handled the request
 * @param url the URL of the request
 * @param status the status code of the request
 * @param time the time it took to handle the request (or true if it was cached)
 */
function log(
  nodeId: string,
  url: string,
  status: number,
  time: number | boolean
) {
  console.log(
    `[node: ${nodeId}] ${url} - ${status} (${
      time === true ? "cached" : time + "ms"
    })`
  );
}

type InfluxLog = {
  nodeId: string;
  url: string;
  status: number;
  time?: number;
  cached?: boolean;
};
/**
 * Creates an InfluxDB point and writes it to the database
 *
 * @param nodeId the node ID that handled the request
 * @param url the URL of the request
 * @param status the status code of the request
 * @param time the time it took to handle the request (or true if it was cached)
 */
async function logRequestToDatabase({
  nodeId,
  url,
  status,
  time,
  cached,
}: InfluxLog) {
  if (totalRequests === undefined) {
    const metrics: any = await MetricsSchema.findOne({ _id: "proxy" }).exec();
    if (metrics) {
      totalRequests = metrics.totalRequests as number;
    } else {
      totalRequests = 0;
    }
  }
  totalRequests++;
  MetricsSchema.updateOne(
    { _id: "proxy" },
    { $set: { totalRequests: totalRequests } },
    { upsert: true }
  ).exec();

  const point = new Point("proxy")
    .tag("type", "request")
    .tag("node", nodeId)
    .booleanField("cached", cached ? true : false)
    .stringField("url", url)
    .intField("status", status)
    .timestamp(Date.now());
  if (time) {
    point.intField("time", time);
  }
  try {
    InfluxWriteApi.writePoint(point);
    InfluxWriteApi.writePoint(
      new Point("proxy")
        .intField("totalRequests", totalRequests)
        .timestamp(Date.now())
    );
  } catch (ex) {
    console.log("Failed to write to influx");
    console.log(ex);
  }
}

export default class ProxyRoute extends Route {
  constructor() {
    super({ path: "/" });
  }

  async handle(req: Request, res: Response) {
    const url = req.url.substring(1);
    if (!url) {
      res.status(400).json(RouteMessages.badRequest("No URL provided"));
      return;
    }
    if (IGNORED_PATHS.includes(url)) {
      res.status(404).send("Not found");
      return;
    }

    try {
      const cachedRequest = cache.get<CachedRequest>(url);
      if (cachedRequest) {
        const { status, headers, data, nodeId } = cachedRequest;

        res.status(status).set(headers).json(data);
        log(nodeId, url, status, true);
        return logRequestToDatabase({
          nodeId: nodeId,
          url,
          status: status,
          cached: true,
        });
      }

      const node = nodeManager.getRandomNode();
      if (!node) {
        res
          .status(500)
          .json(RouteMessages.internalServerError("No nodes available"));
        return;
      }

      const before = Date.now();
      const response = await node.fetch(url);
      const nodeId = response.headers["x-proxy-node"];
      const { status, headers, data } = response;

      if (response.status === 500) {
        res.status(500).json(RouteMessages.internalServerError(data));
        return;
      }

      log(nodeId, url, status, Date.now() - before);
      logRequestToDatabase({
        nodeId,
        url,
        status: status,
        time: Date.now() - before,
        cached: false,
      });
      cache.set(url, {
        nodeId: nodeId,
        status: status,
        headers: headers,
        data: data,
      } as CachedRequest);

      // Send the response to the client
      res.status(status).set(headers).send(data);
    } catch (ex: any) {
      res.status(500).json(RouteMessages.internalServerError(ex.message || ex));
    }
  }
}
